//
// Created by philipp on 30.11.17.
//

#include <iostream>
#include "../include/PlatoonProtocolLib/LeaderVehicle.h"

namespace platoonProtocol
{

using namespace std::chrono_literals;

constexpr networking::time::Duration LeaderVehicle::HEARTBEAT_TIMEOUT;
constexpr networking::time::Duration LeaderVehicle::BROADCAST_INTERVAL;

LeaderVehicle::Ptr LeaderVehicle::create(networking::Networking & net, const NetworkInfo & info)
{
    return std::make_shared<LeaderVehicle>(PrivateTag{}, net, info);
}

LeaderVehicle::LeaderVehicle(PrivateTag, networking::Networking & net, const NetworkInfo & info)
    : Vehicle(net, info, Role::LEADER)
{
    broadcastTimer = networking::time::Timer::create(net);
}

void LeaderVehicle::stop()
{
    Vehicle::stop();

    removeAllFollowers();

    stopBroadcasting();
}

void LeaderVehicle::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    this->platoonConfig = platoonConfig;
}

void LeaderVehicle::setPlatoonSpeed(PlatoonSpeed platoonSpeed)
{
    platoonConfig = PlatoonConfig{platoonSpeed, platoonConfig.get().innerPlatoonDistance};
}

void LeaderVehicle::setInnerPlatoonDistance(InnerPlatoonDistance innerPlatoonDistance)
{
    platoonConfig = PlatoonConfig{platoonConfig.get().platoonSpeed, innerPlatoonDistance};
}

std::size_t LeaderVehicle::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(followersMutex);
    return followers.size();
}

void LeaderVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    Vehicle::doCreatePlatoon();

    // Generate a platoon id.
    platoonId = rand();
}

void LeaderVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    Vehicle::doRunPlatoon();

    startBroadcasting();
}

void LeaderVehicle::doLeavePlatoon()
{
    if (state == State::IDLE)
        return;

    Vehicle::doLeavePlatoon();
}

void LeaderVehicle::receiveMessage(const PlatoonMessage & message)
{
    switch (message.messageType)
    {
        case messageTypes::FV_HEARTBEAT:
            receiveHeartbeat(message);
            break;

        case messageTypes::FV_REQUEST:
            receivePlatoonCreateRequest(message);
            break;

        case messageTypes::FV_LEAVE:
            receiveLeavePlatoonMessage(message);
            break;

        default:
            // ignore rest
            break;
    }
}

void LeaderVehicle::receivePlatoonCreateRequest(const PlatoonMessage & request)
{
    // Always accept request.
    sendMessage(PlatoonMessage::acceptResponse(myInfo.vehicleId, request.srcVehicle, platoonId, utils::keys(followers)),
                3s);
    addFollower(request.srcVehicle);
    doRunPlatoon();
}

bool LeaderVehicle::findFollower(VehicleId id) const
{
    return followers.find(id) != followers.end();
}

void LeaderVehicle::addFollower(VehicleId vehicleId)
{
    if (findFollower(vehicleId))
        return;

    Follower follower;
    follower.vehicleId = vehicleId;

    {
        std::lock_guard<std::mutex> lock(followersMutex);
        followers[vehicleId] = follower;
    }

    startHeartbeatTimer(followers[vehicleId]);
}

void LeaderVehicle::removeFollower(Follower & follower)
{
    stopHeartbeatTimer(follower);

    std::lock_guard<std::mutex> lock(followersMutex);
    followers.erase(follower.vehicleId);
}

void LeaderVehicle::startHeartbeatTimer(Follower & follower)
{
    auto followerId = follower.vehicleId;
    auto self = shared_from_this();
    // Create a new one to make sure we're never getting a busy exception.
    follower.heartbeatTimer = networking::time::Timer::create(net);
    follower.heartbeatTimer->startTimeout(
        HEARTBEAT_TIMEOUT,
        [self, followerId]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            if (!self->findFollower(followerId))
                return;

            log() << "Heartbeat Timeout!\n" << Log::flush;
            auto & follower = self->followers[followerId];
            self->removeFollower(follower);
            if (self->followers.empty())
                self->doLeavePlatoon();
        });
}

void LeaderVehicle::stopHeartbeatTimer(LeaderVehicle::Follower & follower)
{
    follower.heartbeatTimer->stop();
}

void LeaderVehicle::receiveHeartbeat(const PlatoonMessage & heartbeat)
{
    if (!checkFollowerMessage(heartbeat))
        return;

    auto & follower = followers[heartbeat.srcVehicle];
    stopHeartbeatTimer(follower);
    startHeartbeatTimer(follower);
}

void LeaderVehicle::receiveLeavePlatoonMessage(const PlatoonMessage & message)
{
    if (!checkFollowerMessage(message))
        return;

    auto & follower = followers[message.srcVehicle];
    removeFollower(follower);
    if (followers.empty())
        doLeavePlatoon();
}

void LeaderVehicle::startBroadcasting()
{
    sendBroadcastMessage(); // Send broadcast immediately.

    auto self = shared_from_this();
    broadcastTimer->startPeriodicTimeout(
        BROADCAST_INTERVAL,
        [self]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            self->sendBroadcastMessage();
        });
}

void LeaderVehicle::stopBroadcasting()
{
    broadcastTimer->stop();
}

void LeaderVehicle::sendBroadcastMessage()
{
    PlatoonConfig configCopy = platoonConfig.get();
    // Send the broadcast message and forget about it.
    sendMessage(PlatoonMessage::broadcastMessage(
        myInfo.vehicleId,
        platoonId,
        configCopy.platoonSpeed,
        configCopy.innerPlatoonDistance,
        utils::keys(followers)), BROADCAST_INTERVAL);
}

void LeaderVehicle::removeAllFollowers()
{
    utils::foreachInMap(followers,
                        [this](auto & follower)
                        { this->removeFollower(follower); });
}

bool LeaderVehicle::checkFollowerMessage(const PlatoonMessage & message) const
{
    return state == State::RUNNING_PLATOON &&
           message.dstVehicle == myInfo.vehicleId &&
           findFollower(message.srcVehicle) &&
           message.platoonId == platoonId;
}

LeaderVehicle::Ptr LeaderVehicle::shared_from_this()
{
    return std::static_pointer_cast<LeaderVehicle>(Vehicle::shared_from_this());
}

}